package xapi-stdext-encodings

  1. Overview
  2. Docs

Module Encodings.String_validatorSource

Parameters

Signature

Sourceval is_valid : string -> bool

Returns true if and only if the given string is validly-encoded.

Sourceval validate : string -> unit

Raises an encoding error if the given string is not validly-encoded.

Sourceval longest_valid_prefix : string -> string

Returns the longest validly-encoded prefix of the given string.